
/**********************************************************************************
 *                                                                                *
 * dstrds_fused: Factorization routine for Aasen blocked triangular               *
 *                tridiagonalization                                              *
 *                                                                                *
 *                                                                                *
 * See 'driver.c' for a propoer usage example.                                    *
 *                                                                                *
 *                                                                                *
 *  Gil Shklarski and Sivan Toledo, May 2007                                      *
 *                                                                                *
 **********************************************************************************/

#include <math.h>
#include <assert.h>
#include <stdio.h>

#define min(x,y) (((x) < (y))? (x) : (y))
#define GIVENS(X,i,j) { tA0 = (X)[i]; tA1 = (X)[j]; (X)[i] = c*tA0+s*tA1; (X)[j] =-s*tA0+c*tA1; }


/* BLAS Headers */
void dcopy_(int * N, double * DX, int * INCX, double * DY,int * INCY);
void dscal_(int * N, double * DA, double * DX, int *INCX);
void daxpy_(int * N, double * DA, double * DX, int * INCX, double * DY, int * INCY);
void dgemv_(char * TRANS,int * M, int * N,double * ALPHA, double * A,int * LDA,
            double * X, int * INCX, double * BETA, double * Y, int * INCY);
void dgemm_(char * TRANSA, char * TRANSB, int * M,int * N,int *K,
	    double * ALPHA, double * A, int * LDA,
	    double * B, int * LDB, 
	    double * BETA, double * C, int * LDC);
void dger_(int * M, int * N,double * ALPHA,
	   double * X,int * INCX, double *Y,int *INCY,
	   double *A,int *LDA);

int  idamax_(int * N,  double * DX, int * INCX);
void dswap_(int * N, double * DX, int * INCX, double * DY,int * INCY);


/*
   N    = (INPUT) matrix order.
   NB   = (INPUT) number of columns in a partition
   A    = (INPUT/OUTPUT) matrix stored colomnwise
   LDA  = (INPUT) A's leading dimension
   IPIV = (OUTPUT) pivoting index switch array (of size N)
*/
void dstrds_fused( int * N, double * A, int *LDA, int * IPIV, 
		   int * NB, double * B, int * LDB, int * NRHS) {

  
  int n = *N;
  int nb = *NB;
  /* (int)ceil((double)n/(double)(*nb));*/
  int J,i,j;
  int minJplusNBandN;
  double zero = 0.0, one = 1.0, minus_one = -1.0;
  char charN = 'N',charT = 'T';
  int M;
  int one_int = 1, two_int = 2, three_int = 3;
  int lda = *LDA;
  int ldb = *LDB;
  int k;
  double x,y,r,c,s,scale;
  double tA0,tA1;
  int rng1,rng2;
  int flag;

  int I; /*XXX*/

  if (n == 1) {
    scale = 1/A[1];
    dscal_(NRHS,&scale,B+1,LDB);
    return;
  } else if (n == 2) {
    /* Assume there is only one block */
    if (IPIV[1] != 1)
      dswap_(NRHS,B+1,LDB,B+2,LDB);

    for (j=1;j<=(*NRHS);j++) {
      tA0 = B[1+(j-1)*ldb];
      tA1 = B[2+(j-1)*ldb];

      scale = (A[1]*A[4]-A[3]*A[3]);
      B[1+(j-1)*ldb] = (tA0*A[4]-A[3]*tA1)/scale;
      B[2+(j-1)*ldb] = (A[1]*tA1-A[3]*tA0)/scale;
    }
    if (IPIV[1] != 1)
      dswap_(NRHS,B+1,LDB,B+2,LDB);
    return;
  } 


  /* printf("XXX NRHS = %d, LDB = %d, LDA = %d, NB = %d, nb = %d\n", *NRHS, *LDB, *LDA, *NB, nb);*/

  
  for (J=1; J <= n; J+=nb) {
    minJplusNBandN = min(J+nb,n);
    for (i=(J+1); i <= minJplusNBandN; i++) 
      if (i != IPIV[i]) 
	dswap_(NRHS,B+i,LDB,B+IPIV[i],LDB);

    for (i=(J+1); i <= minJplusNBandN; i++) {
      if (i > 2) {
	M = n-i+1;
	/* dgemm_(&charN, &charN, &M, NRHS, &one_int,
	   &minus_one, A+i+(i-2)*lda, LDA,
	   B+i-1, LDB, &one, 
	   B+i, LDB);
	*/
	dger_(&M,NRHS,&minus_one,
	      A+i+(i-2)*lda, &one_int,
	      B+i-1,LDB,
	      B+i,LDB);
      }
      if (i > 3) {
	j = i-3;
	
	c = A[j+1];
	s = A[j+(j+2)*lda];


	for (k=1; k <= (*NRHS); k++)
	  GIVENS(B,j+(k-1)*ldb,j+1+(k-1)*ldb);
      }
    }
  }
  
  c = A[n-1];
  s = A[1+4*lda];
  for (k=1; k <= (*NRHS); k++)
    GIVENS(B,n-2+(k-1)*ldb,n-1+(k-1)*ldb);

  c = A[n];
  s = A[2+5*lda];

  for (k=1; k <= (*NRHS); k++)
    GIVENS(B,n-1+(k-1)*ldb,n+(k-1)*ldb);




  /*
  for (i=n;i >= 3; i--) {
    scale = 1.0/A[i+(i-1)*lda];
    dscal_(NRHS, &scale,B+i,LDB);
    for (j=1; j<=(*NRHS);j++) {
      B[i-2+(j-1)*ldb] -= A[i-2+(i-1)*lda]*B[i + (j-1)*ldb];
      B[i-1+(j-1)*ldb] -= A[i-1+(i-1)*lda]*B[i + (j-1)*ldb];
    }
  }

  {
    scale = 1.0/A[2+lda];
    dscal_(NRHS,&scale,B+2,LDB);
    scale = -A[1+lda];
    daxpy_(NRHS,&scale,B+2,LDB,B+1,LDB);
  }
  scale = 1.0/A[1];
  dscal_(NRHS,&scale,B+1,LDB);
  */
  /*  mexPrintf("N = %d, nb = %d\n",n,nb); */

  /* SOLVE L' XXX */
  J = n;
  flag = 0;
  if ((n % nb) == 0)
    rng1 = (n/nb-1)*nb+1;
  else
    rng1 = (int)floor((double)n/(double)nb)*nb+1;
  rng2 = n;

  while (rng1 >= 1) {
    /*mexPrintf("rng =[ %d %d] \n",rng1,rng2); */

    
    for (i=rng2; i >= rng1; i--) {
      scale = 1.0/A[i+(i-1)*lda];
      dscal_(NRHS,&scale,B+i,LDB);
      if (i > 2) {
	
	scale = -A[i-2+(i-1)*lda];
	daxpy_(NRHS,&scale,B+i,LDB,B+(i-2),LDB);
	scale = -A[i-1+(i-1)*lda];
	daxpy_(NRHS,&scale,B+i,LDB,B+(i-1),LDB);

      } else {
	if (i>1) {
	  scale = -A[1+lda];
	  daxpy_(NRHS,&scale,B+2,LDB,B+1,LDB);
	}
      }
    }
    /*for (k=1;k<=n;k++)
      mexPrintf("%g\n",B[k]);
    */
    if (rng1 ==1) {
      rng1 = 2;
      flag = 1;
    }

    /*
    for (i=n;i>=(rng1+1);i--) {
      // x(rng(1):min((i-1),rng(2)),:) = x(rng(1):min((i-1),rng(2)),:)-L(i,rng(1):min(i-1,rng(2)))'*x(i,:);
      
      M = min(i-1,rng2)-rng1+1;
      
      
      
	      
      //dgemm_(&charT,&charN,&M,NRHS,&one_int,
      // &minus_one, A+i+(rng1-1)*lda, LDA,
      //B+i,LDB, &one,
      //B+rng1, LDB );
	      
      dger_(&M,NRHS,&minus_one,
	    A+i+(rng1-1)*lda,LDA,
	    B+i,LDB,
	    B+rng1, LDB );
    }
    */


    for (i=rng2; i>=rng1; i--) {
      M = n-i;
      if (M > 0)
	dgemv_(&charT,&M,NRHS,&minus_one,
	       B+i+1,LDB,
	       A+i+1+(i-1)*lda,&one_int,
	       &one,B+i,LDB);
    }


    if (flag == 1) {
      rng1 = 1;
      flag = 0;
    }
    

    i=min(rng2+1,n);
    for (;i>=(rng1+1);i--)
      if (i != IPIV[i])
	dswap_(NRHS,B+i,LDB,B+IPIV[i],LDB);
    
    rng2 = rng1-1;
    rng1 = rng1-nb;
    
  }

  
}
