function [L,T,P] = partitioned_parlett_reid_vanila_code(A,NB)
%
% function [L,T,P] = partitioned_parlett_reid_vanila_code(A,NB)
% 
% For a symmetric matrix A, finds [L,T,P] s.t. P*A*P' = L*T*L', where L is lower-triangular, 
% and T is tridiagonal and symmetric. 
%
% NB is the requested number of partitions (defaults to 1).
%
% Performs partitioned 'Parlett-Reid' formulation, with partial pivoting.
%
% This is demonstration plain-vanilla code only.
%
% Gil Shklarski and Sivan Toledo, 30 May 2007
%

n = size(A,1);

if (nargin < 2)
    NB = 1;
end

e1 = zeros(n,1);
e1(1) = 1;

p = 1:n;
L = [e1, zeros(n,n-1)];
T = zeros(n);

nb = ceil(n/NB);
display(sprintf('Partition columns number = %d',nb));

J=0;
for BL=1:NB
    % factorize block BL

    Jb = J+1;
  
    for J=Jb:min(n,J+nb)
        
        % left looking
        A(J:n,J) = A(J:n,J) - A(J:n,Jb:(J-1))*L(J,Jb:(J-1))';
        
        if (J > Jb)
            T(J-1,J) = T(J,J-1);
            tmp = A(J:n,J) - L(J:n,J-1)*T(J-1,J);
        else
            tmp = A(J:n,J);
        end

        T(J,J) = tmp(1); % because L(j,j)=1
        tmp = tmp - L(J:n,J)*T(J,J);

        if (J<n)
            [max_abs_tmp, max_ind] = max(abs(tmp(2:end)));
            max_ind = max_ind + 1;
            max_tmp = tmp(max_ind); 
   
            if ((max_ind ~= 2) && (max_abs_tmp ~=0))
                % symmetric permutation
                m1 = 2;
                m2 = max_ind;

                tmp(m2) = tmp(m1);
                tmp(m1) = max_tmp;

                m1 = m1 + J-1;
                m2 = m2 + J-1;
               
                A([m1 m2],:) = A([m2 m1],:);
                A(:,[m1 m2]) = A(:,[m2 m1]);

                L([m1 m2],:) = L([m2 m1],:);

                tp=p(m1);
                p(m1) = p(m2);
                p(m2) = tp;

            end

            T(J+1,J) = tmp(2);
            if (tmp(2) ~= 0)
                L(J:n,J+1) = tmp / T(J+1,J);
            else
                L(J+1,J+1) = 1;
            end
        end
    end
    
    % right looking blocked updated
    if (J < n)
        A((J+1):n,(J+1):n) = A((J+1):n,(J+1):n) - ...
            [A((J+1):n,Jb:J),L((J+1):n,J)]*[L((J+1):n,Jb:J)' ; L((J+1):n,J+1)'*T(J+1,J)];
        % Plain-vanilla update (updates the complete trailing submatrix)
        T(J,J+1) = T(J+1,J);
    end
    
end

L = tril(L);
P(1:n,p) = eye(n);
