function [L,T] = recursive_parlett_reid(A)
%
% function [L,T] = recursive_parlett_reid(A)
% 
% For a symmetric matrix A, finds [L,T] s.t. A = L*T*L', where L is lower-triangular, 
% and T is tridiagonal and symmetric. 
%
% Performs recursive 'Parlett-Reid' formulation, with no partial pivoting.
%
% This is demonstration plain-vanilla code only.
%
% Gil Shklarski and Sivan Toledo, 30 May 2007
%

n = length(A);

e1 = zeros(n,1); 
e1(1) = 1;

[L,T] = rec_func(A,e1);


%%
% The recursive routine. 
% Gets the first column of L as an extra input. 
function [L,T] = rec_func(A,first_L_column)

n = length(A);

if (length(first_L_column) ~= n)
    error('Invalid input, A and first_L_column should have the same length');
end


if (n == 1)
    L = 1;
    T = A;
    return;
elseif (n == 2)
    L = [first_L_column, [0;1]];
    T = [A(1,1), A(1,2)-L(2,1)*A(1,1); A(1,2)-L(2,1)*A(1,1), A(2,2)+L(2,1)^2*A(1,1)-2*L(2,1)*A(2,1)]; 
    return;
end

m = ceil(n/2);

[L11,T11] = rec_func(A(1:m,1:m),first_L_column(1:m));

A21 = A((m+1):n,1:m);
H21 = (L11\(A21'))';
L21 = zeros(n-m,m);

L21(:,1) = first_L_column((m+1):n);
L21(:,2) = (H21(:,1)-L21(:,1)*T11(1,1)) / T11(2,1);
for j=3:m
    L21(:,j) = (H21(:,j-1)- L21(:,j-2)*T11(j-2,j-1) - L21(:,j-1)*T11(j-1,j-1)) / T11(j,j-1); 
end

next_block_first_L_column = H21(:,m) - L21(:,m-1)*T11(m-1,m) - L21(:,m)*T11(m,m);
t = next_block_first_L_column(1);

A22 = A((m+1):n,(m+1):n);
[L22,T22] = rec_func(A22-H21*L21'- L21(:,m)*next_block_first_L_column', next_block_first_L_column/t);

T12 = zeros(m,n-m);
T12(m,1) = t;
T = [T11, T12; T12' T22];
L = [L11, zeros(m,n-m); L21, L22];
   